# Copyright (c) THL A29 Limited, a Tencent company. All rights reserved

import os
import json
import pickle
import random
import numpy as np

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.utils.data as data

class SynopGround(data.Dataset):

    def __init__(self, split, cfg=None):
        super(SynopGround, self).__init__()
        self.split = split
        self.vfeat_type = cfg["dataset"]["vfeat_type"]
        # data root paths
        self.sfeat_root = "data/features/video_slowfast"
        self.cfeat_root = "data/features/video_clip"
        self.ofeat_root = "data/features/video_ocr"
        self.tfeat_root = f"data/features/text_roberta"
        self.ann_root = "data/annotations/%s.json"%split
        self.samples = json.load(open(self.ann_root, 'r', encoding="utf-8"))
        self.shuffle = True if self.split == 'train' else False
        self.shuffle_prob = 1

    def _align_feat(self, in_feat, tgt_len):
        src_len, feat_dim = in_feat.shape
        out_feat = F.adaptive_avg_pool1d(torch.from_numpy(in_feat).float().transpose(0, 1).contiguous().view(1, feat_dim, src_len), tgt_len)  # [1, feat_dim, tgt_len]
        out_feat = out_feat[0].transpose(0, 1).contiguous()  # [tgt_len, feat_dim]
        return out_feat

    def __getitem__(self, idx):
        sample = self.samples[idx]
        drama_id = sample['basic_info']['drama_id']
        video_id = sample['basic_info']['video_id']
        vlen = sample['basic_info']['video_length']

        # start-end timestamps
        st_eds = [(timestamp['st_time'], timestamp['ed_time']) for timestamp in sample['timestamps']]
        for st, ed in st_eds:
            assert st <= vlen and ed <= vlen and st < ed, (st, ed, vlen)

        # slowfast features
        video_feat = np.load(os.path.join(self.sfeat_root, "%s_%s.npy"%(drama_id, video_id)))
        video_feat = torch.from_numpy(video_feat) if len(video_feat) <= 2500 else self._align_feat(video_feat, 2500)
        # clip features
        clip_feat = np.load(os.path.join(self.cfeat_root, "%s_%s.npy"%(drama_id, video_id)))
        clip_feat = self._align_feat(clip_feat, video_feat.shape[0])  # [len(video_feat), c]
        # ocr features
        ocr_feat = np.load(os.path.join(self.ofeat_root, "%s_%s.npy"%(drama_id, video_id)))
        ocr_feat = self._align_feat(ocr_feat, video_feat.shape[0])  # [len(video_feat), c]
        # concatenated features
        feat_mapping = {'s': video_feat, 'c': clip_feat, 'o': ocr_feat}
        used_feats = [feat_mapping[feat_type] for feat_type in self.vfeat_type]
        video_feat = torch.cat(used_feats, dim=1)

        # tfeat: npara x [1, para_len, hidden_size]
        tfeat = pickle.load(open(os.path.join(self.tfeat_root, "%s_%s.pkl"%(drama_id, video_id)), 'rb'))
        tfeat = [torch.from_numpy(feat[0]) for feat in tfeat] # npara x [para_len, hidden_size]
        tfeat_mask = [torch.ones(feat.shape[0]) for feat in tfeat]  # npara x [para_len]

        # shuffle the order of paragraphs
        assert len(st_eds) == len(tfeat) == len(tfeat_mask)
        if self.shuffle and random.random() < self.shuffle_prob:
            id_list = list(range(len(st_eds)))
            random.shuffle(id_list)
            st_eds = [st_eds[i] for i in id_list]
            tfeat = [tfeat[i] for i in id_list]
            tfeat_mask = [tfeat_mask[i] for i in id_list]

        # pad the text features and text masks
        tfeat = nn.utils.rnn.pad_sequence(tfeat, batch_first=True) # [npara, ntoken_max, hidden_size]
        tfeat_mask = nn.utils.rnn.pad_sequence(tfeat_mask, batch_first=True) # [npara, ntoken_max]

        # convert start-end labels to center-duration labels
        centers = np.array([(st + ed) / 2 for (st, ed) in st_eds]) / vlen
        durations = np.array([ed - st for (st, ed) in st_eds]) / vlen

        # construct the moment mask for the ground-truth temporal interval
        moment_mask = torch.zeros([centers.shape[0], video_feat.shape[0]])  # [npara, len(video_feat)]
        for i, (st_time, ed_time) in enumerate(st_eds):
            st_ind = max(0, round(st_time / vlen * len(video_feat)))
            ed_ind = min(len(video_feat), round(ed_time / vlen * len(video_feat)))

            # corner cases where the normalized span of target moment is too small
            if st_ind == ed_ind:
                if st_ind < len(video_feat):
                    st_ind, ed_ind = st_ind, st_ind + 1
                else:
                    st_ind, ed_ind = ed_ind - 1, ed_ind

            moment_mask[i, st_ind: ed_ind] = 1

        item = {
            'idx': idx,  # sample index
            'center': torch.from_numpy(centers).float(), # [npara]
            'duration': torch.from_numpy(durations).float(), # [npara]
            'ann_mask': torch.ones(centers.shape[0]),  # [npara]
            'video_feat': video_feat, # [vfeat_len, c]
            'video_mask': torch.ones(video_feat.shape[0]),  # [vfeat_len]
            'text_feat': tfeat, # [npara, ntoken_max, hidden_size]
            'text_mask': tfeat_mask, # [npara, ntoken_max]
            'moment_mask': moment_mask, # [npara, vfeat_len]
        }

        return item

    def __len__(self):
        return len(self.samples)

    @staticmethod
    def collate_fn(batch):
        batch_idx = [b['idx'] for b in batch]
        batch_center = [b['center'] for b in batch]
        batch_duration = [b['duration'] for b in batch]
        batch_ann_mask = [b['ann_mask'] for b in batch]
        batch_video_feat = [b['video_feat'] for b in batch]
        batch_video_mask = [b['video_mask'] for b in batch]

        bs = len(batch)
        hidden_size = batch[0]['text_feat'].shape[-1]
        max_npara = max([b['text_mask'].shape[0] for b in batch])
        max_ntoken = max([b['text_mask'].shape[1] for b in batch])
        max_vfeat_len = max([b['moment_mask'].shape[1] for b in batch])

        batch_text_feat = torch.zeros([bs, max_npara, max_ntoken, hidden_size], dtype=batch[0]['text_feat'].dtype)
        batch_text_mask = torch.zeros([bs, max_npara, max_ntoken], dtype=batch[0]['text_mask'].dtype)
        batch_moment_mask = torch.zeros([bs, max_npara, max_vfeat_len], dtype=batch[0]['moment_mask'].dtype)
        
        for i in range(len(batch)):
            text_feat = batch[i]['text_feat']
            text_mask = batch[i]['text_mask']
            moment_mask = batch[i]['moment_mask']
            vfeat_len = moment_mask.shape[1]
            npara, ntoken = text_mask.shape
            batch_text_feat[i, :npara, :ntoken] = text_feat
            batch_text_mask[i, npara:, 0] = 1  # a hack to ensure each row (token dim) has at least one to attend, to avoid all inf attention map
            batch_text_mask[i, :npara, :ntoken] = text_mask
            batch_moment_mask[i, npara:] = 1  # also a hack too avoid loss error
            batch_moment_mask[i, :npara, :vfeat_len] = moment_mask

        batch_data = {
            'idx': batch_idx,
            'center': nn.utils.rnn.pad_sequence(batch_center, batch_first=True), # [b, max_npara]
            'duration': nn.utils.rnn.pad_sequence(batch_duration, batch_first=True), # [b, max_npara]
            'ann_mask': nn.utils.rnn.pad_sequence(batch_ann_mask, batch_first=True), # [b, max_npara]
            'video_feat': nn.utils.rnn.pad_sequence(batch_video_feat), # [max_video_len, b, c]
            'video_mask': nn.utils.rnn.pad_sequence(batch_video_mask, batch_first=True), # [b, max_video_len]
            'text_feat': batch_text_feat, # [b, max_npara, max_ntoken_len, hidden_size]
            'text_mask': batch_text_mask, # [b, max_npara, max_ntoken_len]
            'moment_mask': batch_moment_mask, # [b, max_npara, max_vfeat_len]
        }

        return batch_data