# Copyright (c) THL A29 Limited, a Tencent company. All rights reserved
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

'''
Local-Global Multimodal Reasoner (LGMR) class

Modified from DETR (https://github.com/facebookresearch/detr)
'''

import torch
from torch import nn
import torch.nn.functional as F

from .detr_decoder import DETRDecoder
from .video_encoder import LocalGlobalEncoder
from .query_decoder import LocalGlobalDecoder


class LGMR(nn.Module):
    
    def __init__(self, cfg):
        super(LGMR, self).__init__()
        self.vfeat_dim = sum([cfg["dataset"]["vfeat_map"][ftype] for ftype in cfg["dataset"]["vfeat_type"]])
        self.tfeat_dim = cfg["dataset"]["tfeat_size"]
        self.nhead = cfg["model"]["nhead"]
        self.dropout = cfg["model"]["dropout"]
        self.d_model = cfg["model"]["d_model"]
        self.dim_feedforward = cfg["model"]["dim_feedforward"]
        self.enc_layers = cfg["model"]["enc_layers"]
        self.dec_layers = cfg["model"]["dec_layers"]
        self.nsubpara = cfg["model"]["nsubpara"]
        self.window_size = cfg["model"]["window_size"]

        # feature projection
        self.vfeat_proj = nn.Linear(self.vfeat_dim, self.d_model)
        self.tfeat_proj = nn.Linear(self.tfeat_dim, self.d_model)

        # local-global temporal encoder
        self.video_encoder = LocalGlobalEncoder(num_layers=self.enc_layers,
                                                window_size=self.window_size,
                                                d_model=self.d_model,
                                                dim_feedforward=self.dim_feedforward,
                                                nhead=self.nhead,
                                                dropout=self.dropout)

        # for extract sub-query
        self.sub_query = nn.Parameter(torch.randn(self.nsubpara, self.d_model))
        self.gfc = nn.Linear(self.d_model, self.d_model)
        self.lfc = nn.Linear(self.d_model, self.d_model)
        self.sub_query_norm = nn.LayerNorm(self.d_model)
        self.subquery_generator = DETRDecoder(num_layers=1,
                                              d_model=self.d_model,
                                              nhead=self.nhead,
                                              dim_feedforward=self.dim_feedforward,
                                              dropout=self.dropout)

        # local-global iterative decoder
        self.query_decoder = LocalGlobalDecoder(num_layers=self.dec_layers,
                                                d_model=self.d_model,
                                                nhead=self.nhead,
                                                dim_feedforward=self.dim_feedforward,
                                                dropout=self.dropout)

        # prediction head
        self.pred_layer = MLP(self.d_model, self.d_model, 2, 3)

    def _pos_1d(self, x):
        # 1-d sinusoidal positional encodings
        t, b, c = x.shape
        not_mask = torch.ones([b, t], dtype=x.dtype, device=x.device)
        t_embed = not_mask.cumsum(1, dtype=torch.float32)
        dim_t = torch.arange(c, dtype=torch.float32, device=x.device)
        dim_t = 10000 ** (2 * (dim_t // 2) / c)
        pos = t_embed[:, :, None] / dim_t
        pos = torch.stack((pos[:, :, 0::2].sin(), pos[:, :, 1::2].cos()), dim=3).flatten(2).permute(1, 0, 2).contiguous() # [t, b, c]
        return pos

    def forward(self, sample, return_weights=False):
        # video_padding_mask: [b, vfeat_len], memory: [vfeat_len, b, c], memory_pos: [vfeat_len, b, c]
        video_padding_mask = sample['video_mask'].eq(0)
        video_memory = self.video_encoder(self.vfeat_proj(sample['video_feat']),
                                          src_mask=video_padding_mask)
        video_memory_pos = self._pos_1d(video_memory)

        # encoded_text: [b, npara, ntoken, c], text_mask: [b, npara, ntoken]
        encoded_text = self.tfeat_proj(sample['text_feat'])
        b, npara, ntoken, c = encoded_text.shape

        # global_query: [npara, b, c], global_query_mask: [b, npara]
        global_avg_mask = sample['text_mask'] / sample['text_mask'].sum(2, keepdim=True)  # a hack has ensured at least one token is valid in a paragraph
        global_query = torch.sum(encoded_text * global_avg_mask.unsqueeze(-1), dim=2).transpose(0, 1).contiguous()  # [npara, b, c]
        global_query_padding_mask = sample['ann_mask'].eq(0)  # [b, npara]
        global_pos = self._pos_1d(global_query)  # [npara, b, c]

        # local_query: [k, b*npara, c]
        sub_query = self.sub_query.unsqueeze(1).repeat(1, b, 1)  # [k, b, c]

        # inter-paragraph position encoding
        sub_query_pos = global_pos.transpose(0, 1).contiguous().view(1, b*npara, c).repeat(sub_query.shape[0], 1, 1) # [k, b*npara, c]

        # encode the global query semantics into local queries
        sub_query = sub_query.unsqueeze(2).repeat(1, 1, npara, 1).view(-1, b*npara, c)  # [k, b*npara, c]
        sub_query = self.sub_query_norm(self.gfc(global_query).transpose(0, 1).contiguous().view(1, b*npara, c) + self.lfc(sub_query))  # [k, b*npara, c]

        # prepare text memory features
        text_memory = encoded_text.view(b*npara, ntoken, c).transpose(0, 1).contiguous()  # [ntoken, b*npara, c]
        text_memory_pos = self._pos_1d(text_memory)  # [ntoken, b*npara,  c]
        text_memory_mask = sample['text_mask'].view(-1, ntoken).eq(0)  # [b*npara, ntoken]
        sub_query = self.subquery_generator(query=sub_query, query_mask=None,
                                            memory=text_memory, memory_mask=text_memory_mask,
                                            pos=text_memory_pos, query_pos=sub_query_pos,
                                            return_weights=return_weights)  # [b*npara, k, c]
        if return_weights:
            sub_query, sub_query_weights = sub_query  # [1, b*npara, k, c], [1, b*npara, k, ntoken]
            sub_query_weights = sub_query_weights[-1]  # [b*npara, k, ntoken]

        sub_query = sub_query[-1].transpose(0, 1).contiguous()  # [k, b*npara, c]
        local_query = sub_query  # [k, b*npara, c]
        local_pos = self._pos_1d(local_query)  # [k, b*npara, c]

        dec_out = self.query_decoder(local_query=local_query,
                                     global_query=global_query, global_query_mask=global_query_padding_mask,
                                     memory=video_memory, memory_mask=video_padding_mask,
                                     pos=video_memory_pos, local_pos=local_pos,
                                     global_pos=global_pos, return_weights=return_weights)

        if return_weights:
            dec_out, raw_middle_att_weights, middle_att_weights, global_att_weights = dec_out

        # produce grounding outputs
        ground_out = self.pred_layer(dec_out).sigmoid() # [nlayer, b, npara, 2]

        model_outputs = {}
        model_outputs['ground_out'] = ground_out  # [nlayer, b, npara, 2]

        if return_weights:
            model_outputs['sub_query_weights'] = sub_query_weights  # [1, b*npara, k, ntoken]
            model_outputs['raw_middle_att_weights'] = raw_middle_att_weights  # [nlayer, b*npara, k, vfeat_len]
            model_outputs['middle_att_weights'] = middle_att_weights  # [nlayer, b, npara, vfeat_len]
            model_outputs['global_att_weights'] = global_att_weights  # [nlayer, b, npara, vfeat_len]

        return model_outputs


class MLP(nn.Module):
    """ Very simple multi-layer perceptron (also called FFN)"""

    def __init__(self, input_dim, hidden_dim, output_dim, num_layers):
        super().__init__()
        self.num_layers = num_layers
        h = [hidden_dim] * (num_layers - 1)
        self.layers = nn.ModuleList(nn.Linear(n, k) for n, k in zip([input_dim] + h, h + [output_dim]))

    def forward(self, x):
        for i, layer in enumerate(self.layers):
            x = F.relu(layer(x)) if i < self.num_layers - 1 else layer(x)
        return x