# Copyright (c) THL A29 Limited, a Tencent company. All rights reserved
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

'''
Query Decoder class

Modified from DETR (https://github.com/facebookresearch/detr)
'''

import torch
from torch import nn
from .transformer import TransformerDecoderLayer, CrossAttentionLayer, _get_clones


class LocalGlobalDecoder(nn.Module):

    def __init__(self, num_layers=3, d_model=512, dim_feedforward=2048, nhead=8, dropout=0.1):
        super(LocalGlobalDecoder, self).__init__()
        
        # for intermediate feature normalization
        self.intermediate_norm = nn.LayerNorm(d_model)

        # for local-global interaction
        cross_att_layer = CrossAttentionLayer(d_model=d_model, nhead=nhead, dropout=dropout)

        # for all query decoding
        decoder_layer = TransformerDecoderLayer(d_model=d_model, nhead=nhead, dim_feedforward=dim_feedforward, dropout=dropout, activation='relu')

        # for local query decoding
        self.ldec_layers = _get_clones(decoder_layer, num_layers)

        # for global query decoding
        self.gdec_layers = _get_clones(decoder_layer, num_layers)

        # for local-global interaction
        self.l2g_layers = _get_clones(cross_att_layer, num_layers)
        self.g2l_layers = _get_clones(cross_att_layer, num_layers)

        # for local-global fusion
        self.out_fc = nn.Linear(d_model * 2, d_model)

    def forward(self, local_query, global_query, global_query_mask, memory, memory_mask,
                pos, local_pos, global_pos, return_weights=False):
        # local_query: [k, b*npara, c], global_query: [npara, b, c], global_query_mask: [b, npara]
        # memory: [vfeat_len, b, c], memory_mask: [b, vfeat_len]
        npara, b, c = global_query.shape
        local_output, global_output = local_query, global_query

        intermediate = []
        if return_weights:
            raw_local_weights = []
            local_weights = []
            global_weights = []

        for i in range(len(self.ldec_layers)):
            # for local query
            local_output = self.ldec_layers[i](local_output,  # [k, b*npara, c]
                                               memory.unsqueeze(2).repeat(1, 1, npara, 1).view(-1, b*npara, c),  # [vfeat_len, b*npara, c]
                                               memory_key_padding_mask=memory_mask.unsqueeze(1).repeat(1, npara, 1).view(b*npara, -1),  # [b*npara, vfeat_len]
                                               pos=pos.unsqueeze(2).repeat(1, 1, npara, 1).view(-1, b*npara, c),  # [vfeat_len, b*npara, c]
                                               query_pos=local_pos,  # [k, b*npara, c]
                                               return_weights=return_weights)
            if return_weights:
                # local_att_weights: [b*npara, k, vfeat_len]
                local_output, local_att_weights = local_output
                raw_local_weights.append(local_att_weights)
                local_att_weights = torch.mean(local_att_weights, dim=1).view(b, npara, -1)  # [b, npara, vfeat_len]
                local_weights.append(local_att_weights)  # [b, npara, vfeat_len]

            # for global query
            global_output = self.gdec_layers[i](global_output,  # [npara, b, c]
                                                memory,  # [vfeat_len, b, c]
                                                tgt_key_padding_mask=global_query_mask,  # [b, npara]
                                                memory_key_padding_mask=memory_mask,  # [b, vfeat_len]
                                                pos=pos,
                                                query_pos=global_pos,
                                                return_weights=return_weights)
            if return_weights:
                # global_weights: [b, npara, vfeat_len]
                global_output, global_att_weights = global_output
                global_weights.append(global_att_weights)

            # for local-global interaction
            local_output = self.g2l_layers[i](local_output,  # [k, b*npara, c]
                                              global_output.unsqueeze(2).repeat(1, 1, npara, 1).view(npara, b*npara, c),  # [npara, b*npara, c]
                                              tgt_key_padding_mask=None,
                                              memory_key_padding_mask=global_query_mask.unsqueeze(1).repeat(1, npara, 1).view(b*npara, npara),  # [b*npara, npara]
                                              pos=global_pos.unsqueeze(2).repeat(1, 1, npara, 1).view(npara, b*npara, c),  # [npara, b*npara, c],
                                              query_pos=local_pos,  # [k, b*npara, c]
                                              return_weights=False)
            global_output = self.l2g_layers[i](global_output.transpose(0, 1).contiguous().view(1, b*npara, c),  # [1, b*npara, c]
                                               local_output,  # [k, b*npara, c]
                                               tgt_key_padding_mask=None,
                                               memory_key_padding_mask=None,
                                               pos=None,
                                               query_pos=None,
                                               return_weights=False)
            global_output = global_output.view(b, npara, c).transpose(0, 1).contiguous()  # [npara, b, c]

            # collect intermediate feature outputs
            inter_local_output = torch.mean(local_output, dim=0).view(b, npara, c)  # [k, b*npara, c] -> [b, npara, c]
            inter_global_output = global_output.transpose(0, 1).contiguous()  # [npara, b, c] -> [b, npara, c]
            intermediate.append(self.intermediate_norm(self.out_fc(torch.cat([inter_local_output, inter_global_output], dim=-1))))  # [b, npara, c]

        if return_weights:
            return torch.stack(intermediate), torch.stack(raw_local_weights), torch.stack(local_weights), torch.stack(global_weights)
        
        return torch.stack(intermediate)