# Copyright (c) THL A29 Limited, a Tencent company. All rights reserved
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

'''
Transformer class

Modified from DETR (https://github.com/facebookresearch/detr)
'''

import copy
from typing import Optional

import torch
from torch import nn, Tensor
import torch.nn.functional as F


class TransformerEncoder(nn.Module):

    def __init__(self, encoder_layer, num_layers, norm=None):
        super().__init__()
        self.layers = _get_clones(encoder_layer, num_layers)
        self.num_layers = num_layers
        self.norm = norm

    def forward(self, src,
                mask: Optional[Tensor] = None,
                src_key_padding_mask: Optional[Tensor] = None,
                pos: Optional[Tensor] = None):
        output = src

        for layer in self.layers:
            output = layer(output,
                           src_mask=mask,
                           src_key_padding_mask=src_key_padding_mask,
                           pos=pos)

        if self.norm is not None:
            output = self.norm(output)

        return output


class TransformerDecoder(nn.Module):

    def __init__(self, decoder_layer, num_layers, norm=None, return_intermediate=False):
        super().__init__()
        self.layers = _get_clones(decoder_layer, num_layers)
        self.num_layers = num_layers
        self.norm = norm
        self.return_intermediate = return_intermediate

    def forward(self, tgt, memory,
                tgt_mask: Optional[Tensor] = None,
                memory_mask: Optional[Tensor] = None,
                tgt_key_padding_mask: Optional[Tensor] = None,
                memory_key_padding_mask: Optional[Tensor] = None,
                pos: Optional[Tensor] = None,
                query_pos: Optional[Tensor] = None,
                return_weights=False):
        output = tgt

        intermediate = []
        if return_weights:
            weights = []
            
        for layer in self.layers:
            output = layer(output, memory,
                           tgt_mask=tgt_mask,
                           memory_mask=memory_mask,
                           tgt_key_padding_mask=tgt_key_padding_mask,
                           memory_key_padding_mask=memory_key_padding_mask,
                           pos=pos, query_pos=query_pos,
                           return_weights=return_weights)
            
            if return_weights:
                output, att_weights = output
                
            if self.return_intermediate:
                intermediate_output = self.norm(output) if self.norm is not None else output
                intermediate.append(intermediate_output)
                
                if return_weights:
                    weights.append(att_weights)

        if self.norm is not None:
            output = self.norm(output)

        if self.return_intermediate:
            if return_weights:
                return torch.stack(intermediate), torch.stack(weights)
            return torch.stack(intermediate)

        if return_weights:
            return output.unsqueeze(0), att_weights
        
        return output.unsqueeze(0)


class LocalGlobalEncoderLayer(nn.Module):

    def __init__(self, d_model=512, nhead=8, dim_feedforward=2048, dropout=0.1, window_size=25, activation="relu"):
        super().__init__()
        
        # window size
        self.winsz = window_size
        
        # local attention
        self.self_attn_local = nn.MultiheadAttention(d_model, nhead, dropout=dropout)

        # attention pooling
        self.attn_pool_toq = nn.Linear(d_model, d_model, bias=False)
        self.attn_pool_tok = nn.Linear(d_model, d_model, bias=False)
        self.attn_pool_tov = nn.Linear(d_model, d_model)
        self.attn_pool_norm = nn.LayerNorm(d_model)

        # global attention
        self.self_attn_global = nn.MultiheadAttention(d_model, nhead, dropout=dropout)

        # feed-forward network
        self.linear1 = nn.Linear(d_model, dim_feedforward)
        self.dropout = nn.Dropout(dropout)
        self.linear2 = nn.Linear(dim_feedforward, d_model)

        # normalization layers
        self.norm1 = nn.LayerNorm(d_model)  # local attn norm
        self.norm2 = nn.LayerNorm(d_model)  # global attn norm
        self.norm3 = nn.LayerNorm(d_model)  # ffn norm  

        # dropout layers
        self.dropout1 = nn.Dropout(dropout)  # local attn residual
        self.dropout3 = nn.Dropout(dropout)  # ffn residual

        self.activation = _get_activation_fn(activation)

    def with_pos_embed(self, tensor, pos: Optional[Tensor]):
        return tensor if pos is None else tensor + pos

    def _pos_1d(self, x):
        t, b, c = x.shape
        not_mask = torch.ones([b, t], dtype=x.dtype, device=x.device)
        t_embed = not_mask.cumsum(1, dtype=torch.float32)
        dim_t = torch.arange(c, dtype=torch.float32, device=x.device)
        dim_t = 10000 ** (2 * (dim_t // 2) / c)
        pos = t_embed[:, :, None] / dim_t
        pos = torch.stack((pos[:, :, 0::2].sin(), pos[:, :, 1::2].cos()), dim=3).flatten(2).permute(1, 0, 2).contiguous() # [t, b, c]
        return pos

    def forward_post(self,
                     src,
                     src_mask: Optional[Tensor] = None,
                     src_key_padding_mask: Optional[Tensor] = None,
                     pos: Optional[Tensor] = None):
        assert len(src) % self.winsz == 0
        t, b, c = src.shape

        # local self-attention
        q = k = self.with_pos_embed(src, pos)
        q = q.view(-1, self.winsz, b, c).transpose(0, 1).contiguous().view(self.winsz, -1, c)  # [winsz, nwin*b, hidden_size]
        k = k.view(-1, self.winsz, b, c).transpose(0, 1).contiguous().view(self.winsz, -1, c)  # [winsz, nwin*b, hidden_size]
        src_local = src.view(-1, self.winsz, b, c).transpose(0, 1).contiguous().view(self.winsz, -1, c)  # [winsz, nwin*b, hidden_size]

        if src_mask is not None:
            raise NotImplementedError
        
        if src_key_padding_mask is not None:
            # use .clone() before inplace operation
            mask_local = src_key_padding_mask.clone().view(b, -1, self.winsz).transpose(0, 1).contiguous().view(-1, self.winsz)  # [nwin*b, winsz]
            # this is a hack to avoid nan, i.e., ensure at least one non-padding position for each window
            mask_local[:, 0] = False

        src2 = self.self_attn_local(query=q, key=k, value=src_local, attn_mask=src_mask,
                                    key_padding_mask=mask_local)[0]
        src2 = src2.view(self.winsz, -1, b, c).transpose(0, 1).contiguous().view(t, b, c)  # local features
        src = src + self.dropout1(src2)  # update the input features by adding the local features
        src = self.norm1(src)  # [t, b, c]

        # local-to-global attentive aggregation
        pool_mask = (1 - src_key_padding_mask.float()).view(b, -1, self.winsz).permute(1, 2, 0)[:, :, :, None]  # [nwin, winsz, b, 1]
        src_avg_global = torch.sum(src.view(-1, self.winsz, b, c) * pool_mask, dim=1) / torch.sum(pool_mask, dim=1).clip(min=1)  # [nwin, b, c]
        # [b*nwin, 1, c] x [b*nwin, c, winsz] -> [b*nwin, 1, winsz] -> [b, nwin, winsz]
        global_to_local_logits = torch.bmm(self.attn_pool_toq(src_avg_global).permute(1, 0, 2).contiguous().view(-1, 1, c),
                                           self.attn_pool_tok(src).view(-1, self.winsz, b, c).permute(2, 0, 1, 3).contiguous().view(-1, self.winsz, c).transpose(1, 2).contiguous())
        global_to_local_logits = global_to_local_logits.view(b, -1, self.winsz)  # [b, nwin, winsz]
        global_to_local_attns = torch.softmax(global_to_local_logits + (src_key_padding_mask.view(b, -1, self.winsz) * -1e10), dim=-1)
        # [b*nwin, 1, winsz] x [b*nwin, winsz, c] -> [b*nwin, 1, c] -> [b, nwin, c] -> [nwin, b, c]
        src_global = torch.bmm(global_to_local_attns.view(-1, 1, self.winsz),
                               self.attn_pool_tov(src).view(-1, self.winsz, b, c).permute(2, 0, 1, 3).contiguous().view(-1, self.winsz, c))
        # [b*nwin, 1, c] -> [b, nwin, c] -> [nwin, b, c]
        src_global = src_global.view(b, -1, c).permute(1, 0, 2).contiguous()
        src_global = self.attn_pool_norm(src_global + src_avg_global)  # [nwin, b, c]
        q = k = self.with_pos_embed(src_global, self._pos_1d(src_global))  # [nwin, b, c]

        # global self-attention
        if src_key_padding_mask is not None:
            # if a window has all padded positions, then it is padded
            mask_global = (src_key_padding_mask.view(b, -1, self.winsz).float().sum(-1) == self.winsz)

        src2 = self.self_attn_global(query=q, key=k, value=src_global, attn_mask=src_mask,
                                     key_padding_mask=mask_global)[0]
        src2 = src2[:, None, :, :].repeat(1, self.winsz, 1, 1).view(-1, b, c)
        src = src + src2
        src = self.norm2(src)

        # FFN
        src2 = self.linear2(self.dropout(self.activation(self.linear1(src))))
        src = src + self.dropout3(src2)
        src = self.norm3(src)
        return src

    def forward(self,
                src,
                src_mask: Optional[Tensor] = None,
                src_key_padding_mask: Optional[Tensor] = None,
                pos: Optional[Tensor] = None):

        return self.forward_post(src, src_mask, src_key_padding_mask, pos)


class TransformerDecoderLayer(nn.Module):

    def __init__(self, d_model, nhead, dim_feedforward=2048, dropout=0.1, activation="relu"):
        super().__init__()
        
        # self-attention and cross-attention
        self.self_attn = nn.MultiheadAttention(d_model, nhead, dropout=dropout)
        self.multihead_attn = nn.MultiheadAttention(d_model, nhead, dropout=dropout)
        
        # feed-forward network
        self.linear1 = nn.Linear(d_model, dim_feedforward)
        self.dropout = nn.Dropout(dropout)
        self.linear2 = nn.Linear(dim_feedforward, d_model)
        
        # layer normalization
        self.norm1 = nn.LayerNorm(d_model)
        self.norm2 = nn.LayerNorm(d_model)
        self.norm3 = nn.LayerNorm(d_model)
        
        # dropout
        self.dropout1 = nn.Dropout(dropout)
        self.dropout2 = nn.Dropout(dropout)
        self.dropout3 = nn.Dropout(dropout)

        self.activation = _get_activation_fn(activation)

    def with_pos_embed(self, tensor, pos: Optional[Tensor]):
        return tensor if pos is None else tensor + pos

    def forward_post(self, tgt, memory,
                     tgt_mask: Optional[Tensor] = None,
                     memory_mask: Optional[Tensor] = None,
                     tgt_key_padding_mask: Optional[Tensor] = None,
                     memory_key_padding_mask: Optional[Tensor] = None,
                     pos: Optional[Tensor] = None,
                     query_pos: Optional[Tensor] = None,
                     return_weights=False):
        
        # self-attention
        q = k = self.with_pos_embed(tgt, query_pos)
        tgt2 = self.self_attn(query=q, key=k, value=tgt, attn_mask=tgt_mask,
                              key_padding_mask=tgt_key_padding_mask)[0]
        tgt = tgt + self.dropout1(tgt2)
        tgt = self.norm1(tgt)
        
        # cross-attention
        tgt2, att_weights = self.multihead_attn(query=self.with_pos_embed(tgt, query_pos),
                                                key=self.with_pos_embed(memory, pos),
                                                value=memory, attn_mask=memory_mask,
                                                key_padding_mask=memory_key_padding_mask)
        tgt = tgt + self.dropout2(tgt2)
        tgt = self.norm2(tgt)
        
        # feed-forward network
        tgt2 = self.linear2(self.dropout(self.activation(self.linear1(tgt))))
        tgt = tgt + self.dropout3(tgt2)
        tgt = self.norm3(tgt)
        
        if return_weights:
            return tgt, att_weights
        
        return tgt

    def forward(self, tgt, memory,
                tgt_mask: Optional[Tensor] = None,
                memory_mask: Optional[Tensor] = None,
                tgt_key_padding_mask: Optional[Tensor] = None,
                memory_key_padding_mask: Optional[Tensor] = None,
                pos: Optional[Tensor] = None,
                query_pos: Optional[Tensor] = None,
                return_weights=False):

        return self.forward_post(tgt, memory,
                                 tgt_mask, memory_mask,
                                 tgt_key_padding_mask, memory_key_padding_mask,
                                 pos, query_pos,
                                 return_weights)


class CrossAttentionLayer(nn.Module):

    def __init__(self, d_model, nhead, dropout=0.1):
        super().__init__()
        
        # cross-attention
        self.multihead_attn = nn.MultiheadAttention(d_model, nhead, dropout=dropout)
        
        # layer normalization
        self.norm = nn.LayerNorm(d_model)
        
        # dropout
        self.dropout = nn.Dropout(dropout)

    def with_pos_embed(self, tensor, pos: Optional[Tensor]):
        return tensor if pos is None else tensor + pos

    def forward_post(self, tgt, memory,
                     tgt_mask: Optional[Tensor] = None,
                     memory_mask: Optional[Tensor] = None,
                     tgt_key_padding_mask: Optional[Tensor] = None,
                     memory_key_padding_mask: Optional[Tensor] = None,
                     pos: Optional[Tensor] = None,
                     query_pos: Optional[Tensor] = None,
                     return_weights=False):
        
        # cross-attention only
        tgt2, att_weights = self.multihead_attn(query=self.with_pos_embed(tgt, query_pos),
                                                key=self.with_pos_embed(memory, pos),
                                                value=memory, attn_mask=memory_mask,
                                                key_padding_mask=memory_key_padding_mask)
        tgt = tgt + self.dropout(tgt2)
        tgt = self.norm(tgt)
        
        if return_weights:
            return tgt, att_weights
        
        return tgt

    def forward(self, tgt, memory,
                tgt_mask: Optional[Tensor] = None,
                memory_mask: Optional[Tensor] = None,
                tgt_key_padding_mask: Optional[Tensor] = None,
                memory_key_padding_mask: Optional[Tensor] = None,
                pos: Optional[Tensor] = None,
                query_pos: Optional[Tensor] = None,
                return_weights=False):
            
        return self.forward_post(tgt, memory,
                                 tgt_mask, memory_mask,
                                 tgt_key_padding_mask, memory_key_padding_mask,
                                 pos, query_pos,
                                 return_weights)

def _get_clones(module, N):
    return nn.ModuleList([copy.deepcopy(module) for i in range(N)])

def _get_activation_fn(activation):
    """Return an activation function given a string"""
    if activation == "relu":
        return F.relu
    if activation == "gelu":
        return F.gelu
    if activation == "glu":
        return F.glu
    raise RuntimeError(F"activation should be relu/gelu, not {activation}.")