# Copyright (c) THL A29 Limited, a Tencent company. All rights reserved
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

'''
Video Encoder class

Modified from DETR (https://github.com/facebookresearch/detr)
'''

import torch
from torch import nn
from .transformer import TransformerEncoder, LocalGlobalEncoderLayer


class LocalGlobalEncoder(nn.Module):

    def __init__(self, num_layers=2, window_size=25, d_model=512, dim_feedforward=2048, nhead=8, dropout=0.1):
        super(LocalGlobalEncoder, self).__init__()
        self.window_size = window_size
        
        encoder_layer = LocalGlobalEncoderLayer(d_model=d_model,
                                                window_size=window_size,
                                                nhead=nhead,
                                                dim_feedforward=dim_feedforward,
                                                dropout=dropout,
                                                activation="relu")

        self.encoder = TransformerEncoder(encoder_layer, num_layers=num_layers)

    def _pos_1d(self, x):
        # 1-d sinusoidal positional encodings
        t, b, c = x.shape
        not_mask = torch.ones([b, t], dtype=x.dtype, device=x.device)
        t_embed = not_mask.cumsum(1, dtype=torch.float32)
        dim_t = torch.arange(c, dtype=torch.float32, device=x.device)
        dim_t = 10000 ** (2 * (dim_t // 2) / c)
        pos = t_embed[:, :, None] / dim_t
        pos = torch.stack((pos[:, :, 0::2].sin(), pos[:, :, 1::2].cos()), dim=3).flatten(2).permute(1,0,2).contiguous() # [t, b, c]
        return pos

    def forward(self, x, src_mask):
        # x: [t, b, c], src_mask: [b, t]
        t, b, c = x.shape

        # padding for local window partition
        if t % self.window_size != 0:
            to_pad = self.window_size - t % self.window_size
            x = nn.functional.pad(x, (0, 0, 0, 0, 0, to_pad), 'constant', 0)  # pad zeros to the first dimension
            src_mask = nn.functional.pad(src_mask, (0, to_pad), 'constant', True)  # pad True to the last dimension

        pos = self._pos_1d(x)  # [t_pad, b, c]
        out = self.encoder(src=x, src_key_padding_mask=src_mask, pos=pos) # [t_pad, b, c]

        return out[:t]  # [t, b, c]