# Copyright (c) THL A29 Limited, a Tencent company. All rights reserved

import os
import sys
sys.path.insert(0, os.path.join(os.path.dirname(__file__), ".."))

import pickle
import numpy as np
from tqdm import tqdm

import torch
from torch.utils.data import DataLoader
torch.autograd.set_detect_anomaly(True)

import argparse
import datasets
from models import build_model
from utils.utils import load_config, sample_to_cuda, eval

def inference(model, dataloader):
    all_sids, all_preds = [], []
    all_video_masks, all_moment_masks = [], []
    all_ann_masks, all_text_masks = [], []
    all_sub_query_weights = []
    all_global_att_weights = []

    with torch.no_grad():
        model.eval()
        ious = []
        
        for sample in tqdm(dataloader):
            sample = sample_to_cuda(sample)
            model_outputs = model(sample, return_weights=True)
            pred = model_outputs["ground_out"]  # [nlayer, b, npara, 2]
            sub_query_weights = model_outputs["sub_query_weights"]  # [1, b*npara, k, ntoken]
            global_att_weights = model_outputs["global_att_weights"]  # [nlayer, b, npara, vfeat_len]
            target = torch.stack([sample["center"], sample["duration"]], -1)  # [b, npara, 2]
            b, npara, ntoken, _ = sample["text_feat"].shape
            
            def split_and_convert(input_tensor):
                return list(map(lambda x: x.squeeze(0).cpu().numpy(), torch.split(input_tensor, 1)))
            
            all_sids.extend(sample["idx"])
            all_preds.extend(split_and_convert(pred[-1]))  # [b, npara, 2]
            all_video_masks.extend(split_and_convert(sample["video_mask"]))  # [b, vfeat_len]
            all_moment_masks.extend(split_and_convert(sample["moment_mask"]))  # [b, npara, vfeat_len]
            all_ann_masks.extend(split_and_convert(sample["ann_mask"]))  # [b, npara]
            all_text_masks.extend(split_and_convert(sample["text_mask"]))  # [b, npara, ntoken]
            all_sub_query_weights.extend(split_and_convert(sub_query_weights.view(b, npara, -1, ntoken)))  # [b, npara, k, ntoken]
            all_global_att_weights.extend(split_and_convert(global_att_weights[-1]))  # [b, npara, vfeat_len]

            ious.append(eval(pred[-1], target, sample["ann_mask"]))  # [#valid_preds]
        
        iou_np = np.concatenate(ious, 0)  # flattened iou values from all samples are concatenated into one long sequence
        mIoU = iou_np.mean()

    print('Evaluation mIoU: %.4f'%(mIoU))

    pstr = ""
    for thres in [0.1, 0.3, 0.5, 0.7]:
        pstr += "IoU@%s: %.4f,"%(str(thres), (iou_np > thres).mean())
    print(pstr)

    results_to_save = [all_sids, all_preds, all_video_masks, all_moment_masks, all_ann_masks, all_text_masks, all_sub_query_weights, all_global_att_weights]

    return results_to_save

def main():

    # exp config
    parser = argparse.ArgumentParser()
    parser.add_argument("--ckpt", required=True, type=str)
    parser.add_argument("--cfg", default="configs/base_config.yml", type=str)

    # parse arguments
    args = parser.parse_args()

    # load base config
    config = load_config(args.cfg)
    print(config)

    model = build_model(config).cuda()

    model_ckpt_ddp = torch.load(args.ckpt)
    model_ckpt = {k.replace("module.", ""): v for k, v in model_ckpt_ddp.items()}
    
    model.load_state_dict(model_ckpt)

    dset_name = config["dataset"]["dset_name"]
    testset = getattr(datasets, dset_name)(split="test", cfg=config)
    test_loader = DataLoader(testset,
                             batch_size=config["val"]["batch_size"],
                             shuffle=False,
                             drop_last=False,
                             num_workers=8,
                             pin_memory=True,
                             collate_fn=testset.collate_fn)

    results_to_save = inference(model, test_loader)

    save_path = os.path.join(os.path.dirname(args.ckpt), os.path.basename(args.ckpt).split(".")[0] + "_inference_results.pkl")
    pickle.dump(results_to_save, open(save_path, "wb"))

    print("save results to %s"%save_path)

if __name__ == "__main__":
    main()