# Copyright (c) THL A29 Limited, a Tencent company. All rights reserved

import os
import sys
sys.path.insert(0, os.path.join(os.path.dirname(__file__), ".."))

import numpy as np
from tqdm import tqdm

import torch
import torch.distributed as dist
from torch.utils.data import DataLoader
from torch.nn.parallel import DistributedDataParallel as DDP
torch.autograd.set_detect_anomaly(True)

import argparse
import datasets
from models import build_model
from utils.loss import smooth_l1_loss, giou_loss, guided_att_loss
from utils.utils import seed_everything, load_config, sample_to_cuda, loss_to_str, eval

def train_one_epoch(model, dataloader, optimizer, epoch, cfg):
    # set up training
    model.train()
    # set up loss container
    all_loss = {"l1":[], "giou":[], "global_att": []}

    ious = []
    optimizer.zero_grad()
    for i, sample in tqdm(enumerate(dataloader)):

        # convert data to cuda
        sample = sample_to_cuda(sample)

        # model forward
        model_outputs = model(sample, return_weights=True)

        # grounding predictions
        pred = model_outputs["ground_out"]  # [nlayer, b, npara, 2]
        # cross-attention weights
        global_att_weights = model_outputs["global_att_weights"]  # [nlayer, b, npara, vfeat_len]
        # prepare timestamp labels
        target = torch.stack([sample["center"], sample["duration"]], -1)  # [b, npara, 2]

        # compute iou results
        ious.append(eval(pred[-1].detach(), target.detach(), sample["ann_mask"]))  # [#valid_preds]

        # layer-wise decode loss
        dec_layers = cfg["model"]["dec_layers"]
        loss_l1, loss_giou, loss_global_att = 0.0, 0.0, 0.0

        for j in range(dec_layers):
            loss_l1 += smooth_l1_loss(pred[j], target, sample["ann_mask"]) / dec_layers
            loss_giou += giou_loss(pred[j], target, sample["ann_mask"]) / dec_layers
            loss_global_att += guided_att_loss(global_att_weights[j], sample["moment_mask"], sample["ann_mask"]) / dec_layers

        # total loss
        loss = loss_l1 + cfg["train"]["lambda1"] * loss_giou
        loss = loss + cfg["train"]["lambda2"] * loss_global_att

        # scaling loss value for gradient accumulation
        loss /= cfg["train"]["iter_size"]

        # update the loss container
        all_loss["l1"].append(loss_l1.data.cpu().numpy())
        all_loss["giou"].append(loss_giou.data.cpu().numpy())
        all_loss["global_att"].append(loss_global_att.data.cpu().numpy())

        # back-propagation
        loss.backward()

        # gradient accumulation: update parameters every iter_size times
        if (i + 1) % cfg["train"]["iter_size"] == 0:
            torch.nn.utils.clip_grad_norm_(model.parameters(), cfg["train"]["clip_grad_norm"])
            optimizer.step()
            optimizer.zero_grad()

    # mIoU for current epoch
    mIoU = np.concatenate(ious, 0).mean()

    # loss values for current epoch
    loss_str = loss_to_str(all_loss)

    # display training statistics
    print("Training Statistics: Epoch:%d, %s mIoU=%.4f"%(epoch, loss_str, mIoU))

def evaluate(model, dataloader, epoch, cfg):

    with torch.no_grad():
        # set up evaluation
        model.eval()
        ious = []
        all_loss = {"l1": [], "giou": [], "global_att": []}

        for sample in tqdm(dataloader):
            sample = sample_to_cuda(sample)
            model_outputs = model(sample, return_weights=True)
            pred = model_outputs["ground_out"]  # [nlayer, b, npara, 2]
            global_att_weights = model_outputs["global_att_weights"]  # [nlayer, b, npara, 2]
            target = torch.stack([sample["center"], sample["duration"]], -1)  # [nlayer, b, npara, 2]

            # layer-wise decode loss
            dec_layers = cfg["model"]["dec_layers"]
            loss_l1, loss_giou, loss_global_att = 0.0, 0.0, 0.0

            for j in range(dec_layers):
                loss_l1 += smooth_l1_loss(pred[j], target, sample["ann_mask"]) / dec_layers
                loss_giou += giou_loss(pred[j], target, sample["ann_mask"]) / dec_layers
                loss_global_att += guided_att_loss(global_att_weights[j], sample["moment_mask"], sample["ann_mask"]) / dec_layers

            all_loss["l1"].append(loss_l1.data.cpu().numpy())
            all_loss["giou"].append(loss_giou.data.cpu().numpy())
            all_loss["global_att"].append(loss_global_att.data.cpu().numpy())

            ious.append(eval(pred[-1], target, sample["ann_mask"]))

        iou_np = np.concatenate(ious, 0)
        mIoU = iou_np.mean()
        loss_str = loss_to_str(all_loss)

    print('Evaluate Epoch %d:, %s mIoU:%.4f'%(epoch, loss_str, mIoU))
    pstr = ""

    # compute the recall metrics
    for thres in [0.1, 0.3, 0.5, 0.7]:
        pstr += "IoU@%s: %.4f,"%(str(thres), (iou_np > thres).mean())

    # display the recall metrics
    print(pstr)

    return all_loss, mIoU, pstr

def main():
    print("PID: %s"%(os.getpid()))
    dist.init_process_group(backend="nccl")

    # exp config
    parser = argparse.ArgumentParser()
    parser.add_argument("--exp_id", required=True, type=str)
    parser.add_argument("--local_rank", default=-1, type=int)
    parser.add_argument("--cfg", default="configs/base_config.yml", type=str)

    # parse arguments
    args = parser.parse_args()

    # load base config
    config = load_config(args.cfg)
    print(config)

    # set up local device id
    local_rank = args.local_rank
    torch.cuda.set_device(local_rank)

    # set up random seed
    seed = 1234 + dist.get_rank()
    seed_everything(seed)

    # build up model
    model = build_model(config).to(local_rank)
    model = DDP(model, device_ids=[local_rank], output_device=local_rank, find_unused_parameters=False)

    # build up optimizer
    optimizer = torch.optim.Adam(model.parameters(), lr=config["train"]["lr"])

    # build up dataset
    dset_name = config["dataset"]["dset_name"]
    trainset = getattr(datasets, dset_name)(split="train", cfg=config)
    train_sampler = torch.utils.data.distributed.DistributedSampler(trainset)
    batch_sampler_train = torch.utils.data.BatchSampler(train_sampler, config["train"]["batch_size"], drop_last=True)
    train_loader = DataLoader(trainset,
                              batch_sampler=batch_sampler_train,
                              num_workers=8,
                              pin_memory=True,
                              collate_fn=trainset.collate_fn)

    valset = getattr(datasets, dset_name)(split="val", cfg=config)
    val_loader = DataLoader(valset,
                            batch_size=config["val"]["batch_size"],
                            shuffle=False,
                            drop_last=False,
                            num_workers=8,
                            pin_memory=True,
                            collate_fn=valset.collate_fn)

    testset = getattr(datasets, dset_name)(split="test", cfg=config)
    test_loader = DataLoader(testset,
                             batch_size=config["val"]["batch_size"],
                             shuffle=False,
                             drop_last=False,
                             num_workers=8,
                             pin_memory=True,
                             collate_fn=testset.collate_fn)

    # start the training loop
    epochs, shuffle_epoch = config["train"]["nepoch"], config["train"]["shuffle_epoch"]
    max_val_iou, best_val_log, best_test_log = -1, None, None

    for epoch in range(epochs):
        print("Training Epoch: ", epoch)

        # set random seeds for distributed training
        train_loader.batch_sampler.sampler.set_epoch(epoch)
        # set probabilities of data augmentation
        shuffle_prob = max(1 - epoch / shuffle_epoch) if shuffle_epoch > 0 else 0
        train_loader.dataset.shuffle_prob = shuffle_prob

        # do training for one epoch
        train_one_epoch(model, train_loader, optimizer, epoch=epoch, cfg=config)

        # skip evaluation of early epochs to speed up training
        skip_eval_epoch = -1
        
        if dist.get_rank() == 0 and epoch > skip_eval_epoch:
            # evaluate validation performance
            val_log = evaluate(model, val_loader, epoch=epoch, cfg=config)
            val_miou = val_log[1]

            # check update
            if val_miou > max_val_iou:
                max_val_iou = val_miou
                best_val_log = val_log

                # test the best model
                best_test_log = evaluate(model, test_loader, epoch=epoch, cfg=config)

                # save the best checkpoint
                save_dir = os.path.join("results", args.exp_id)
                if not os.path.isdir(save_dir):
                    os.makedirs(save_dir)
                torch.save(model.state_dict(), os.path.join(save_dir, "best.pth"))

            # display current validation metrics
            print("[Epoch: %d Val]"%epoch, "mIoU: " + str(val_log[1]) + " " + val_log[2])  # print the mIoU and R@m metrics
            
            # display metrics
            print("[Best Val]: ", "mIoU: " + str(best_val_log[1]) + " " + best_val_log[2])  # print the mIoU and R@m metrics
            print("[Best Test]: ", "mIoU: " + str(best_test_log[1]) + " " + best_test_log[2])  # print the mIoU and R@m metrics

if __name__ == "__main__":
    main()