# Copyright (c) THL A29 Limited, a Tencent company. All rights reserved

import torch
import torch.nn.functional as F

def center_to_sted(arr):
    # arr: [*, 2]
    st = arr[..., 0] - arr[..., 1] / 2
    ed = arr[..., 0] + arr[..., 1] / 2
    return torch.stack([st, ed], dim=-1)

def smooth_l1_loss(pred, target, mask):
    # pred, target: [b, npara, 2], mask: [b, npara]
    loss_raw = F.smooth_l1_loss(pred, target, reduction='none').sum(-1)  # [b, npara]
    loss = loss_raw[mask.eq(1)].mean()

    return loss

def giou_loss(pred, target, mask):
    # pred, target: [b, npara, 2], mask: [b, npara]
    pred_ = center_to_sted(pred)
    target_ = center_to_sted(target)

    max_st = torch.max(pred_[:, :, 0], target_[:, :, 0])
    min_ed = torch.min(pred_[:, :, 1], target_[:, :, 1])
    inter = (min_ed - max_st).clamp(min=0)
    union = pred[:, :, 1] + target[:, :, 1] - inter
    
    min_st = torch.min(pred_[:, :, 0], target_[:, :, 0])
    max_ed = torch.max(pred_[:, :, 1], target_[:, :, 1])
    enclosure = max_ed - min_st

    iou = inter / union
    giou = iou - (enclosure - union) / enclosure # [b, npara]
    loss_giou = 1 - giou

    loss_giou = loss_giou[mask.eq(1)].mean()

    return loss_giou

def guided_att_loss(att, gt, mask):
    # att: [b, npara, vfeat_len], gt: [b, npara, vfeat_len], mask: [b, npara]
    valid_att = att[mask.bool()]  # [#valid_preds, vfeat_len]
    valid_gt = gt[mask.bool()]  # [#valid_preds, vfeat_len]

    loss = - torch.log((valid_att * valid_gt).sum(-1).clip(min=1e-5))
    loss = torch.mean(loss)

    return loss