# Copyright (c) THL A29 Limited, a Tencent company. All rights reserved

import os
import yaml
import random

import torch
import numpy as np

def load_config(cfg=None):
    config_file = "configs/base_config.yml" if cfg is None else cfg
    return yaml.load(open(config_file, "r"), Loader=yaml.FullLoader)

def sample_to_cuda(sample):
    for key in sample:
        if key != "idx":
            sample[key] = sample[key].cuda()
    return sample

def loss_to_str(all_loss):
    out_str = ""
    for loss_name in all_loss:
        out_str += "loss_%s: %.4f, "%(loss_name, np.array(all_loss[loss_name]).mean())
    return out_str

def eval(pred, gt, mask):
    # pred: [b, npara, 2], gt: [b, npara, 2], mask: [b, npara]
    pred_valid = pred[mask == 1]
    gt_valid = gt[mask == 1]

    # convert center-duration outputs to start-end outputs
    def center_to_sted(arr):
        # arr: [n, 2]
        return torch.clip(arr[:, 0] - arr[:, 1] / 2, min=0), torch.clip(arr[:, 0] + arr[:, 1] / 2, max=1)

    def cal_ious(pred, gt):
        # pred: [n, 2], gt: [n, 2]
        # ct-dr to st-ed
        pred_s, pred_e = center_to_sted(pred)
        gt_s, gt_e = center_to_sted(gt)
        # iou calculation
        max_start = torch.stack([pred_s, gt_s], -1).max(-1).values
        min_end = torch.stack([pred_e, gt_e], -1).min(-1).values
        inter = torch.clip(min_end - max_start, min=0)
        return inter / (gt[:, 1] + pred[:, 1] - inter)  # [n]

    return cal_ious(pred_valid, gt_valid).cpu().numpy()

def seed_everything(seed):
    os.environ["CUBLAS_WORKSPACE_CONFIG"]=":16:8" # For cuda >= 10.2
    random.seed(seed)
    os.environ["PYTHONHASHSEED"] = str(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False